from moai.monads.object_pose.bpnp import BPnP
from moai.monads.object_pose.epnp import EPnP
from moai.monads.object_pose.rotational_conversions import AngleAxis

__all__ = [
    "BPnP",
    "AngleAxis",
    "EPnP",
]


from moai.monads.geometry.project import Projection
from moai.monads.geometry.transform_points import Transformation
from moai.monads.geometry.transform_ops import (
    Relative,
    Inverse,
    Transpose
)

__all__ = [
    "Projection",
    "Transformation",
    "Relative",
    "Inverse",
    "Transpose",
]


from moai.data.datasets.spherical.ThreeD60.common import (
    _rotate_image,
    _rotate_normal_image,
    ImageTypes
)

import torch
import random
import typing

__all__ = ["RotationDataset"]

#TODO: extract this to augmentation layer
class RotationDataset(torch.utils.data.Dataset):
    def __init__(self, super_dataset):
        super(RotationDataset, self).__init__()
        self.super = super_dataset
        
    def __len__(self) -> int:
        return len(self.super)

    def __getitem__(self, index: int) -> typing.Dict[str, torch.Tensor]:
        entries = self.super[index]
        width = entries['left_down']['color'].shape[2]
        idx = random.randint(0, width - 1)
        for placement, type_map in entries.items():
            if 'color' in type_map:
                entries[placement]['color'] = _rotate_image(entries[placement]['color'], idx)
            if 'depth' in type_map:
                entries[placement]['depth'] = _rotate_image(entries[placement]['depth'], idx)
            if 'normal' in type_map:
                entries[placement]['normal'] = _rotate_normal_image(entries[placement]['normal'], idx)
        return entries

    



import torch
import hydra.utils as hyu
import omegaconf.omegaconf
import inspect
import typing
import itertools
import toolz
import logging

log = logging.getLogger(__name__)

#TODO: use class param execs

__all__ = ["Cascade"]

class Cascade(torch.nn.ModuleDict): #TODO: check if x: ['arg'] is the same as x: 'arg'
    execs: typing.List[typing.Callable]

    def __init__(self, 
        monads: omegaconf.DictConfig,
        **kwargs: typing.Mapping[str, typing.Any]
    ):
        super(Cascade, self).__init__()
        #TODO: first construct via monads and then create lambdas via kwargs
        loop = ((key, params) for key, params in kwargs.items() if key in monads.keys())
        #TODO: add message in case key is not found
        self.execs = []
        for k, p in loop:
            #TODO: if instantiation fails, error out with message saying about adding it to the config file
            self.add_module(k, hyu.instantiate(getattr(monads, k))) #TODO: add construction loop first in class execs, and then funcs based on kwargs
            last_module = toolz.last(self.modules()) #NOTE: moduledict is ordered            
            sig = inspect.signature(last_module.forward)
            props = [prop for prop in sig.parameters if p[prop] is not None]
            for keys in zip(*list(p[prop] for prop in itertools.chain(props, ['out']))):
                self.execs.append(lambda tensor_dict, k=keys, p=props, f=last_module:
                    tensor_dict.update({
                        k[-1]: f(**dict(zip(p, 
                            list(tensor_dict[i] if type(i) is str
                                else list(tensor_dict[j] for j in i) if i is not None else None
                                for i in k[:-1]
                            )
                        )))
                    })
                )                

    def forward(self,
        tensors: typing.Dict[str, torch.Tensor]    
    ) -> typing.Dict[str, torch.Tensor]:
        for exe in self.execs:
            exe(tensors) #NOTE: each executing lambda updates the tensor dict itself
        return tensors



import torch
import logging
import kornia as kn

log = logging.getLogger(__name__)

__all__ = ["AngleAxis"]

#TODO: make_from, convert_to
class AngleAxis(torch.nn.Module):
    def __init__(self,
        convert_to:    str = "rot_mat", #"quat" , "ortho6d", "euler_angles" , etc.        
    ):
        super(AngleAxis,self).__init__()
        if convert_to == "rot_mat":
            self.convert_func = kn.geometry.conversions.angle_axis_to_rotation_matrix
        elif convert_to == "quat":
            #(x, y, z, w) format of quaternion
            self.convert_func = kn.geometry.conversions.angle_axis_to_quaternion
        else:
            log.error(f"The selected rotational convertion {convert_to} is not supported. Please enter a valid one to continue.")

    def forward(self,angle_axis: torch.Tensor) -> torch.Tensor:
        return self.convert_func(angle_axis)


from moai.data.datasets.object_pose.UAVA.importers import (
    load_image,
    load_depth,
    load_normal,
)

from PIL import Image

import torch
import os
import sys
import torch
import json
import logging
import typing

log = logging.getLogger(__name__)

__all__ = ["UAVA"]

class UAVA(torch.utils.data.Dataset):
    def __init__(self,
        root:               str, # path to root folder
        metadata:           str, # path to metadata .json
        split:              str, # [train,test,val]
        drones:             typing.List[str],
        views:              typing.List[str],
        frames:             typing.List[str],
        types:              typing.List[str],
    ):
        super(UAVA,self).__init__()
        self.drones = drones
        self.views = views
        self.frames = frames
        self.types = types
        self.data = {}
        buildings = self.get_splits(split, root)
        if not os.path.exists(metadata):
            error_message = f"Metadata annotations file ({metadata}) not found."
            log.error(error_message)
            raise ValueError(error_message)
        with open(metadata) as f:
            self.annotations = json.load(f)
        for drone in os.listdir(root):
            if drone not in drones:
                log.warning(f"Skipping {drone} drone model as it is not included in the selected drones list.")
                continue
            drone_path = os.path.join(root, drone)
            for view in os.listdir(drone_path):
                if view not in views:
                    log.warning(f"Skipping {view} view as it is not included in the selected views list.")
                    continue
                view_path = os.path.join(drone_path, view)
                for data_type in os.listdir(view_path):
                    if data_type not in types:
                        log.warning(f"Skipping {data_type} data type as it is not included in the selected data types list.")
                        continue
                    for building in buildings:
                        img_path = os.path.join(view_path, data_type, building)
                        for img in os.listdir(img_path):
                            if (len(img.split("_")) <= 6):
                                row, date , frame , __ , view , _ = img.split("_")
                            else:
                                row, date , frame , __ , _, view , ___ = img.split("_")                            
                            if view not in views:
                                continue                            
                            if int(frame) not in frames:
                                continue
                            full_img_name = os.path.join(img_path, img)                            
                            unique_name = building + "_" + str(row) + "_" + str(date)
                            try:
                                if self.annotations['drones'][drone][unique_name]['exocentric']["0"]['visibleAll'] == False:
                                    continue
                            except KeyError:
                                continue
                            if unique_name not in self.data:
                                self.data[unique_name] = {}
                            if view not in self.data[unique_name]:
                                self.data[unique_name][view] = {}
                            if frame not in self.data[unique_name][view]:
                                self.data[unique_name][view][frame] = {}
                            self.data[unique_name][view][frame][data_type] = full_img_name
 
    def __len__(self) -> int:
        return len(self.data)
    
    def __getitem__(self, idx: int) -> typing.Dict[str, torch.Tensor]:
        key = list(self.data.keys())[idx]
        datum = self.data[key]
        datum_out = {}
        for view in self.views:
            for frame in self.frames:
                for data_type in self.types:
                    if data_type == "colour":
                        value = load_image(datum[view][str(frame)][data_type])
                    elif data_type == "silhouette" and view == "exocentric":
                        value = load_image(datum[view][str(frame)][data_type])
                    elif data_type == "colour" or data_type == "silhouette":
                          continue
                    elif data_type == "depth":
                        value = load_depth(
                            datum[view][str(frame)][data_type]
                        )
                    elif data_type == "normal":
                        value = load_normal(
                            datum[view][str(frame)][data_type]
                        )
                    else:
                        value = load_image(
                            datum[view][str(frame)][data_type]
                        )
                    datum_out.update({
                        str(view) + "_"+ str(frame) + "_" + str(data_type): value.squeeze(0)
                    })
                    if view == "exocentric":
                        pose = torch.FloatTensor(json.JSONDecoder().decode(
                            self.annotations['drones'][self.drones[0]][key][view][str(frame)]['pose']
                        )).reshape(4,4)
                        pose_inv = pose.inverse()
                        bbox_2d = torch.FloatTensor(json.JSONDecoder().decode(
                            self.annotations['drones'][self.drones[0]][key][view][str(frame)]['2dbbox']
                        ))
                        bbox_3d = torch.FloatTensor(json.JSONDecoder().decode(
                            self.annotations['drones'][self.drones[0]][key][view][str(frame)]['3dbbox']
                        )).reshape(8,2)
                        points_3d = torch.FloatTensor(json.JSONDecoder().decode(
                            self.annotations['drones'][self.drones[0]][key][view][str(frame)]['3dpoints']
                        )).reshape(9,2)
                        key_root = str(view) + "_"+ str(frame) + "_" 
                        datum_out.update({key_root + "pose" : pose})
                        datum_out.update({key_root + "pose_inv" : pose_inv})
                        datum_out.update({key_root + "2dbbox" : bbox_2d})
                        datum_out.update({key_root + "3dbbox" : bbox_3d})
                        datum_out.update({key_root + "points_3d" : points_3d})
                        datum_out.update({key_root + "key" : key})
                        datum_out.update({key_root + "translation" : pose[:3,3]})
                        datum_out.update({key_root + "rotation" : pose[:3,:3]})
                    else:
                        src = 0 if int(frame) == 1 else 1
                        ego_pose = torch.FloatTensor(json.JSONDecoder().decode(
                            self.annotations['drones'][self.drones[0]][key][view][str(src)]['pose']
                        )).reshape(4,4)
                        key_root = str(view) + "_"+ str(frame) + "_" 
                        datum_out.update({key_root + "key" : key})
                        datum_out.update({key_root + 'source_to_target' : ego_pose})
        return datum_out

    def get_splits(self, split: str, root: str) -> typing.List[str]:
        if split == "train":
            txt_path = os.path.join(root, "data splits", "scenes_train.txt")
        elif split == "test":
            txt_path = os.path.join(root, "data splits", "scenes_test.txt")
        elif split == "val":
            txt_path = os.path.join(root, "data splits", "scenes_val.txt")
        with open(txt_path,'r') as txt_file:
            splits = [line.strip() for line in txt_file]
        return splits



from moai import __version__ as miV
from moai.utils.engine import NoOp as NoInterval
from moai.parameters.optimization import NoOp as NoOptimization
from moai.parameters.optimization.schedule import NoOp as NoScheduling
from moai.parameters.initialization import Default as NoInit
from moai.validation import NoOp as NoValidation
from moai.supervision import NoOp as NoSupervision

import torch
import numpy
import pytorch_lightning
import torch.utils.data as ptd
import omegaconf.omegaconf 
import hydra.utils as hyu
import typing
import types
import toolz
import logging

log = logging.getLogger(__name__)

__all__ = ['FeedForward']

def _create_supervision_block(
    cfg: omegaconf.DictConfig,
    force: bool=True
):
    if force and not cfg:
        log.warning("Empty supervision block in feedforward model.")
    return hyu.instantiate(cfg) if cfg else NoSupervision()

def _create_validation_block(
    cfg: omegaconf.DictConfig,
    force: bool=True
):
    if force and not cfg:
        log.warning("Empty validation block in feedforward model.")
    return hyu.instantiate(cfg) if cfg else NoValidation()

def _create_processing_block(
    cfg: omegaconf.DictConfig, 
    attribute: str, 
    monads: omegaconf.DictConfig,
    force: bool=False
):
    if force and not cfg and attribute in cfg:
        log.warning(f"Empty processing block ({attribute}) in feedforward model.")
    return hyu.instantiate(getattr(cfg, attribute), monads)\
        if cfg and attribute in cfg else torch.nn.Identity()

def _create_interval_block(
    cfg: omegaconf.DictConfig,
    force: bool=False
):
    if force and not cfg:
        log.warning("Empty interval block in feedforward model.")
    return hyu.instantiate(cfg) if cfg else NoInterval()

def _create_optimization_block(
    cfg: omegaconf.DictConfig,
    params: typing.Union[typing.Iterable[torch.Tensor], typing.Dict[str, torch.Tensor]],
    force: bool=False
):
    if force and not cfg:
        log.warning("No optimizer in feedforward model.")
    return hyu.instantiate(cfg, params) if cfg else NoOptimization(params)

def _create_scheduling_block(
    cfg: omegaconf.DictConfig, 
    optimizers: typing.Sequence[torch.optim.Optimizer],
    force: bool=False
):
    if force and not cfg:
        log.warning("No scheduling used in feedforward model.")
    return hyu.instantiate(cfg, optimizers) if cfg else NoScheduling(optimizers)

class FeedForward(pytorch_lightning.LightningModule):
    def __init__(self, 
        data:               omegaconf.DictConfig=None,
        parameters:         omegaconf.DictConfig=None,
        feedforward:        omegaconf.DictConfig=None,
        monads:             omegaconf.DictConfig=None,        
        supervision:        omegaconf.DictConfig=None,
        validation:         omegaconf.DictConfig=None,        
        visualization:      omegaconf.DictConfig=None,
        export:             omegaconf.DictConfig=None,
        hyperparameters:    typing.Union[omegaconf.DictConfig, typing.Mapping[str, typing.Any]]=None,
    ):
        super(FeedForward, self).__init__()        
        self.data = data
        self.initializer = parameters.initialization if parameters is not None else None
        self.optimization_config = parameters.optimization if parameters is not None else None
        self.schedule_config = parameters.schedule if parameters is not None else None
        self.supervision = _create_supervision_block(supervision)
        self.validation = _create_validation_block(validation) #TODO: change this, "empty processing block" is confusing
        self.preprocess = _create_processing_block(feedforward, "preprocess", monads=monads)
        self.postprocess = _create_processing_block(feedforward, "postprocess", monads=monads)
        self.visualizer = _create_interval_block(visualization)
        self.exporter = _create_interval_block(export)        
        #NOTE: __NEEDED__ for loading checkpoint
        hparams = hyperparameters if hyperparameters is not None else { }
        hparams.update({'moai_version': miV})
        self.hparams =  hparams
        self.global_test_step = 0

    def initialize_parameters(self) -> None:
        init = hyu.instantiate(self.initializer) if self.initializer else NoInit()
        init(self)

    def forward(self,
        tensors:                typing.Dict[str, torch.Tensor]
    ) -> typing.Dict[str, torch.Tensor]:
        pass

    def training_step(self, 
        batch:                  typing.Dict[str, torch.Tensor],
        batch_idx:              int,
        optimizer_idx:          int=None,
    ) -> typing.Dict[str, typing.Union[torch.Tensor, typing.Dict[str, torch.Tensor]]]:
        preprocessed = self.preprocess(batch)
        prediction = self(preprocessed)
        postprocessed = self.postprocess(prediction)
        total_loss, losses = self.supervision(postprocessed)
        #TODO: should add loss maps as return type to be able to forward them for visualization
        losses = toolz.keymap(lambda k: f"train_{k}", losses)
        losses.update({'total_loss': total_loss})        
        self.log_dict(losses, prog_bar=False, logger=True)        
        return { 'loss': total_loss, 'tensors': postprocessed }

    def training_step_end(self, 
        train_outputs: typing.Dict[str, typing.Union[torch.Tensor, typing.Dict[str, torch.Tensor]]]
    ) -> None:
        if self.global_step and (self.global_step % self.visualizer.interval == 0):
            self.visualizer(train_outputs['tensors'])
        if self.global_step and (self.global_step % self.exporter.interval == 0):
            self.exporter(train_outputs['tensors'])
        return train_outputs['loss']

    def validation_step(self,
        batch: typing.Dict[str, torch.Tensor],
        batch_nb: int
    ) -> dict:
        preprocessed = self.preprocess(batch)
        prediction = self(preprocessed)
        outputs = self.postprocess(prediction)
        #TODO: consider adding loss maps in the tensor dict
        metrics = self.validation(outputs)
        return metrics

    def validation_epoch_end(self,
        outputs: typing.List[dict]
    ) -> None:
        keys = next(iter(outputs), { }).keys()
        metrics = { }
        for key in keys:
            metrics[key] = numpy.mean(numpy.array(
                [d[key].item() for d in outputs]
            ))        
        self.log_dict(metrics, prog_bar=True, logger=False, on_epoch=True, sync_dist=True)
        log_metrics = toolz.keymap(lambda k: f"val_{k}", metrics)
        self.log_dict(log_metrics, prog_bar=False, logger=True, on_epoch=True, sync_dist=True)
    
    def test_step(self, 
        batch: typing.Dict[str, torch.Tensor],
        batch_nb: int
    ) -> dict:
        preprocessed = self.preprocess(batch)
        prediction = self(preprocessed)
        outputs = self.postprocess(prediction)
        metrics = self.validation(outputs)
        self.global_test_step += 1
        log_metrics = toolz.keymap(lambda k: f"test_{k}", metrics)
        self.log_dict(log_metrics, prog_bar=False, logger=True, on_step=True, on_epoch=False, sync_dist=True)
        return metrics, outputs

    def test_step_end(self,
        metrics_tensors: typing.Tuple[typing.Dict[str, torch.Tensor], typing.Dict[str, torch.Tensor]],        
    ) -> None:
        metrics, tensors = metrics_tensors
        if self.global_test_step and (self.global_test_step % self.exporter.interval == 0):
            self.exporter(tensors)
        if self.global_test_step and (self.global_test_step % self.visualizer.interval == 0):
            self.visualizer(tensors)
        return metrics

    def test_epoch_end(self, 
        outputs: typing.List[dict]
    ) -> dict:
        keys = next(iter(outputs), { }).keys()
        metrics = { }
        for key in keys:
            metrics[key] = numpy.mean(numpy.array(
                [d[key].item() for d in outputs]
            ))   
        self.log_dict(metrics, prog_bar=False, logger=True, on_epoch=True, sync_dist=True)

    def configure_optimizers(self) -> typing.Tuple[typing.List[torch.optim.Optimizer], typing.List[torch.optim.lr_scheduler._LRScheduler]]:
        log.info(f"Configuring optimizer and scheduler")
        self.optimization = _create_optimization_block(self.optimization_config, self.parameters())
        self.schedule = _create_scheduling_block(self.schedule_config, self.optimization.optimizers)
        return self.optimization.optimizers, self.schedule.schedulers

    def train_dataloader(self) -> ptd.DataLoader:        
        log.info(f"Instantiating ({self.data.train.iterator._target_.split('.')[-1]}) train set data iterator")
        train_iterator = hyu.instantiate(self.data.train.iterator)
        train_loader = hyu.instantiate(self.data.train.loader, train_iterator)
        return train_loader

    def val_dataloader(self) -> ptd.DataLoader:        
        log.info(f"Instantiating ({self.data.val.iterator._target_.split('.')[-1]}) validation set data iterator")
        val_iterator = hyu.instantiate(self.data.val.iterator)
        validation_loader = hyu.instantiate(self.data.val.loader, val_iterator)
        return validation_loader

    def test_dataloader(self) -> ptd.DataLoader:        
        log.info(f"Instantiating ({self.data.test.iterator._target_.split('.')[-1]}) test set data iterator")
        test_iterator = hyu.instantiate(self.data.test.iterator)
        test_loader = hyu.instantiate(self.data.test.loader, test_iterator)        
        return test_loader



import re

__all__ = ['sort_nicely']

def tryint(s):
    try:
        return int(s)
    except:
        return s

def alphanum_key(s):
    """ Turn a string into a list of string and number chunks.
        "z23a" -> ["z", 23, "a"]
    """
    return [ tryint(c) for c in re.split('([0-9]+)', s) ]

def sort_nicely(l):
    """ Sort the given list in the way that humans expect.
    """
    l.sort(key=alphanum_key)


from moai.data.datasets.human_pose.HUMAN4D.importers.extrinsics import (
    load_extrinsics,
)
from moai.data.datasets.human_pose.HUMAN4D.importers.intrinsics import (
    load_intrinsics_repository,
    get_intrinsics,
)
from moai.data.datasets.human_pose.HUMAN4D.importers.image import (
    load_depth,
    load_depth_pgm,
    load_image,
)
from moai.data.datasets.human_pose.HUMAN4D.importers.joints import (
    load_skel3D_bin,
)
from moai.data.datasets.human_pose.HUMAN4D.importers.offsets import (
    load_rgbd_skip,
)


import torch

class L2(torch.nn.Module):
    def __init__(self):
        super(L2, self).__init__()

    def forward(self,
        gt: torch.Tensor,
        pred: torch.Tensor,
        weights: torch.Tensor=None, # float tensor
        mask: torch.Tensor=None, # byte tensor
    ) -> torch.Tensor:
        l2 = ((gt - pred) ** 2)
        if weights is not None:
            l2 = l2 * weights
        if mask is not None:
            l2 = l2[mask]        
        return l2


import numpy
import torch

__all__ = ['load_extrinsics']

def load_extrinsics(
    filename: str,
    data_type=torch.float32,
    translation_scale: float=1e-03,
):
    data = numpy.loadtxt(filename)
    pose = numpy.zeros([4, 4])
    pose[3, 3] = 1
    pose[:3, :3] = data[:3, :3]
    pose[:3, 3] = data[3, :] * translation_scale
    #y = R*x + t
    rotation = torch.tensor(data[:3, :3], dtype=data_type)
    translation = torch.tensor(data[3, :], dtype=data_type) * translation_scale
    #y = Rx + t <-> x = Rinv*y - Rinv * t 
    rotation_inverse = torch.tensor(numpy.linalg.inv(pose[:3, :3]), dtype=data_type)
    translation_inverse = torch.tensor(numpy.inner(rotation_inverse, pose[:3, 3]) * (-1), dtype=data_type)
    #Continue after walk
    return rotation, translation, rotation_inverse, translation_inverse


import torch

#NOTE: https://discuss.pytorch.org/t/multi-class-cross-entropy-loss-and-softmax-in-pytorch/24920/4

__all__ = ["CrossEntropy"]

#TODO: to be refactored into a generic multi-/binary classification loss (/w logits, logs or probs in gt/pred)
class CrossEntropy(torch.nn.CrossEntropyLoss):
    def __init__(self,
    
    ):
        super(CrossEntropy, self).__init__()

    def forward(self, 
        gt: torch.Tensor, # class ids
        pred: torch.Tensor, # logits (i.e. raw predictions, no softmax applied -- see link above)
    ) -> torch.Tensor:
        return super(CrossEntropy, self).forward(pred, gt)


def load_rgbd_skip(filename: str, folder: str):
    f = open(filename)
    f_lines = f.readlines()
    skip = 0
    for line in f_lines:
        if (folder in line):
            skip = int(line.split('\t')[1])
            break
    return skip


from moai.validation.metrics.common.rmse import RMSE
from moai.validation.metrics.common.rmsle import RMSLE
from moai.validation.metrics.common.absrel import AbsRel
from moai.validation.metrics.common.sqrel import SqRel

__all__ = [
    "RMSE",
    "RMSLE",
    "AbsRel",
    "SqRel",
]


from moai.data.datasets.spherical.ThreeD60.matterport3d import Matterport3D
from moai.data.datasets.spherical.ThreeD60.stanford2d3d import Stanford2D3D
from moai.data.datasets.spherical.ThreeD60.suncg import SunCG
from moai.data.datasets.spherical.ThreeD60.rotation import RotationDataset
from moai.data.datasets.spherical.ThreeD60.common import (
    Placements,
    ImageTypes,
    extract_image,
    extract_path
)

import typing
import torch

__all__ = ["ThreeD60"]

_dataset_generators = {
    "suncg": lambda *params: SunCG(params[0], params[1], params[2]),
    "m3d": lambda *params: Matterport3D(params[0], params[1], params[2]),
    "s2d3d": lambda *params: Stanford2D3D(params[0], params[1], params[2])
}

def _get_datasets(filename, datasets=["suncg", "s2d3d", "m3d"],              \
        placements=[Placements.CENTER, Placements.RIGHT, Placements.UP],    \
        image_types=[ImageTypes.COLOR, ImageTypes.DEPTH, ImageTypes.NORMAL], 
        longitudinal_rotation = False):
    return torch.utils.data.ConcatDataset(list(map(                                                      \
        lambda d: RotationDataset(_dataset_generators[d](filename, placements, image_types)) 
                if longitudinal_rotation else _dataset_generators[d](filename, placements, image_types), datasets)))

class ThreeD60(torch.utils.data.Dataset):
    def __init__(self,
        filename: str,
        datasets: typing.Sequence[str]=["suncg", "s2d3d", "m3d"],
        placements: typing.Sequence[str]=[Placements.CENTER, Placements.RIGHT, Placements.UP],
        image_types: typing.Sequence[str]=[ImageTypes.COLOR, ImageTypes.DEPTH, ImageTypes.NORMAL],
        longitudinal_rotation: bool=False
    ):
        super(ThreeD60, self).__init__()
        enum_placements = list(map(lambda p: 
            Placements[p.upper()], placements)
        )
        enum_image_types = list(map(lambda t: 
            ImageTypes[t.upper()], image_types)
        )
        self.iterator = _get_datasets(filename, datasets, 
            enum_placements, enum_image_types, longitudinal_rotation
        )

    def __len__(self) -> int:
        return len(self.iterator)

    def __getitem__(self, index) -> typing.Dict[str, torch.Tensor]:
        return self.iterator[index]


import cv2
import torch
import numpy

__all__ = [
    'load_image',
    'load_depth',
    'load_normal',
]

def load_image(
    filename: str, 
    data_type: torch.TensorType=torch.float32,
) -> torch.Tensor:
    color_img = numpy.array(cv2.imread(filename, cv2.IMREAD_ANYCOLOR))
    h, w, c = color_img.shape
    color_data = color_img.astype(numpy.float32).transpose(2, 0, 1)
    return torch.from_numpy(
        color_data.reshape(1, c, h, w)        
    ).type(data_type) / 255.0

def load_depth(
    filename: str,
    data_type: torch.TensorType=torch.float32, 
    scale: float=1.0,
) -> torch.Tensor:
    depth_img = numpy.array(cv2.imread(filename, cv2.IMREAD_ANYDEPTH))
    h, w = depth_img.shape
    depth_data = depth_img.astype(numpy.float32) * scale
    return torch.from_numpy(
        depth_data.reshape(1, 1, h, w)        
    ).type(data_type)

def load_normal(
    filename: str,
    data_type: torch.TensorType=torch.float32
) -> torch.Tensor:
    normal_img = numpy.array(cv2.imread(filename, cv2.IMREAD_UNCHANGED)).transpose(2, 0, 1)
    c, h, w = normal_img.shape
    return torch.from_numpy(normal_img).reshape(1,c, h, w).type(data_type)


from moai.data.datasets.human_pose.HUMAN4D.importers import (
	load_intrinsics_repository,
	get_intrinsics,
	load_extrinsics,
	load_depth_pgm,
	load_image,
	load_skel3D_bin,
	load_rgbd_skip,
)
from moai.data.datasets.human_pose.HUMAN4D.utils import sort_nicely
from moai.data.datasets.human_pose.HUMAN4D.enums import joint_selection
from torch.utils.data import Dataset
from torch.utils.data import DataLoader

import os
import sys
import torch
import numpy
import logging

'''
Dataset importer. We assume that data follows the below structure.
root_path
----|device_repository.json
----|Subject1
	|pose
	----|Action 1
		----|Dump
			----|color
			----|depth
			----|gposes3d
		|
		|
	----|Action N
		----|Dump
			----|color
			----|depth
			----|gposes3d
	|
	|
----SubjectN
'''

log = logging.getLogger(__name__)

__all__ = ["HUMAN4D"]

class HUMAN4D(Dataset):
	def __init__(self, 
		path: str,
		data_types,
		center_crop,
		device_list,
		metadata_path,
		mode,
		depth_threshold,
		clip_frames
	):
		super(HUMAN4D, self).__init__()
		self.center_crop = center_crop
		self.device_list = device_list
		self.depth_threshold = depth_threshold
		self.mode = mode
		device_repo_path = os.path.join(path, "device_repository.json")
		self.data_types = data_types
		if not os.path.exists(device_repo_path):
			raise ValueError("{} does not exist, exiting.".format(device_repo_path))
		self.depth_intrinsics_repository = load_intrinsics_repository(device_repo_path, "depth")
		self.color_intrinsics_repository = load_intrinsics_repository(device_repo_path, "color")

		if not os.path.exists(path):
			raise ValueError("{} does not exist, exiting.".format(path))
		
		use_metadata = True
		if not os.path.exists(os.path.join(path, metadata_path)):
			use_metadata = False
			
		if use_metadata:
			metadata_samples = open(os.path.join(path, metadata_path), 'r').readlines()
			metadata_dict = {}
			for line in metadata_samples:
				values = line.split(' ')
				values[-1] = values[-1].strip()
				if values[1] not in metadata_dict.keys():
					metadata_dict[values[1]] = {}
				if values[2] not in metadata_dict[values[1]].keys():
					metadata_dict[values[1]][values[2]] = []
				metadata_dict[values[1]][values[2]].append(int(values[0]))

		self.data = {}
		self.extrinsics_paths_per_subject = {}
		subject_folders = [subject for subject in os.listdir(path) if "." not in subject and len(subject) == 2]
		# iterate over each recorded folder
		for subject in subject_folders:
			if use_metadata and subject not in metadata_dict.keys():
				continue

			abs_subject_path = os.path.join(path, subject)
			if not os.path.isdir(abs_subject_path):
				continue			

			subject_recordings = [sub_rec for sub_rec in os.listdir(abs_subject_path) if '-' in sub_rec and '!' not in sub_rec and '.' not in sub_rec]
			# We remove and hold separately the pose folder for the Calibration
			# subject_recordings.pop(subject_recordings.index("pose"))
			pose_folder = os.path.join(abs_subject_path, "pose")
			if not os.path.exists(pose_folder):
				log.warning("Folder {} does not contain \"pose\" folder".format(abs_subject_path))
			
			for recording in subject_recordings:
				if use_metadata and recording not in metadata_dict[subject].keys():
					continue	

				recording_modalities = os.path.join(abs_subject_path, recording, "Dump")
				skel_3d_gt_directory = os.path.join(recording_modalities, "gposes3d")				
				
				skel_3d = os.listdir(skel_3d_gt_directory)
				sort_nicely(skel_3d)
				skipped = load_rgbd_skip(os.path.join(abs_subject_path, "offsets.txt"), recording)

				if "color" in data_types:
					color_directory = os.path.join(recording_modalities, "color")
					color_img_paths = os.listdir(color_directory)
					sort_nicely(color_img_paths)
					color_img_paths = color_img_paths[4 * (skipped + 1):]


					for color_img_path in color_img_paths:
						full_filename = os.path.join(color_directory, color_img_path)
						_, ext = os.path.splitext(full_filename)
						
						if ext != ".png":
							continue
						
						if self.mode == "single":
							_id, _name,_type,_ = color_img_path.split("_")

							if use_metadata and int(_id) not in metadata_dict[subject][recording]:
								continue

							unique_name = subject + "-" + recording + "-" + _name + "-" + str(int(_id) - skipped)

							#skip names that we dp now want to load
							if _name not in device_list:
								continue
							
							if unique_name not in self.data:
								self.data[unique_name] = {}						
							
							if "pose" not in self.data[unique_name]:
								self.data[unique_name]['pose'] = os.path.join(pose_folder, _name + ".extrinsics")

							if "dev_name" not in self.data[unique_name]:
								self.data[unique_name]['dev_name'] = _name

							self.data[unique_name][_type] = full_filename
							
						else:
							_id, _name, _type, _ = color_img_path.split("_")
							if use_metadata and int(_id) not in metadata_dict[subject][recording]:
								continue
							unique_name = subject + "-" + recording + "-" + str(int(_id) - skipped)

							#skip names that we dp now want to load
							if _name not in device_list:
								continue
							
							if unique_name not in self.data:
								self.data[unique_name] = {}						

							if _name not in self.data[unique_name]:
								self.data[unique_name][_name] = {}
							
							if "pose" not in self.data[unique_name]:
								self.data[unique_name][_name]['pose'] = os.path.join(pose_folder, _name + ".extrinsics")
							
							self.data[unique_name][_name][_type] = full_filename
				
				if "depth" in data_types:
					depth_directory = os.path.join(recording_modalities, "depth")
					depth_img_paths = os.listdir(depth_directory)
					sort_nicely(depth_img_paths)
					depth_img_paths = depth_img_paths[4 * (skipped + 1):]
					for depth_img_path in depth_img_paths:
						full_filename = os.path.join(depth_directory, depth_img_path)
						_, ext = os.path.splitext(full_filename)
						
						if ext != ".pgm":
							continue

						if self.mode == "single":
							_id, _name, _type,_ = depth_img_path.split("_")
							if use_metadata and int(_id) not in metadata_dict[subject][recording]:								
								continue

							unique_name = subject + "-" + recording + "-" + _name + "-" + str(int(_id) - skipped)
							#skip names that we dp now want to load
							if _name not in device_list:
								continue
							
							if unique_name not in self.data:
								self.data[unique_name] = {}						
							
							if "pose" not in self.data[unique_name]:
								self.data[unique_name]['pose'] = os.path.join(pose_folder, _name + ".extrinsics")
							
							if "dev_name" not in self.data[unique_name]:
								self.data[unique_name]['dev_name'] = _name

							self.data[unique_name][_type] = full_filename
						else:
							_id, _name,_type,_ = depth_img_path.split("_")
							if use_metadata and int(_id) not in metadata_dict[subject][recording]:
								continue
							unique_name = subject + "-" + recording + "-" + str(int(_id) - skipped)
							#skip names that we dp now want to load
							if _name not in device_list:
								continue
							
							if unique_name not in self.data:
								self.data[unique_name] = {}						

							if _name not in self.data[unique_name]:
								self.data[unique_name][_name] = {}
							
							if "pose" not in self.data[unique_name]:
								self.data[unique_name][_name]['pose'] = os.path.join(pose_folder, _name + ".extrinsics")
		
							self.data[unique_name][_name][_type] = full_filename
				#Always load the GT data. 
				### Keep the unique name consistent in all loading scenarios				
				for skel_3d_frame in skel_3d:
					full_filename = os.path.join(skel_3d_gt_directory, skel_3d_frame)
					_id, ext = os.path.splitext(skel_3d_frame)
					if ext != ".npy":
						continue
					if self.mode == "single":
						for _name in self.device_list:
							unique_name = subject + "-" + recording + "-" + _name + "-" + _id
							# this check also satisfies that we get data only from the selected frames in "metadata_dict"
							if unique_name not in self.data:
								continue
							if "joints_3d_raw" not in self.data[unique_name]:
								self.data[unique_name]["joints_3d_raw"] = full_filename
					else:
						unique_name = subject + "-" + recording + "-" + _id

						if unique_name not in self.data:
							continue

						if "joints_3d_raw" not in self.data[unique_name]:
							self.data[unique_name]["joints_3d_raw"] = full_filename				
		log.info(metadata_path.split('.')[0] + " is succesfully loaded.")
			

	def __len__(self):
		return len(self.data)

	def __getitem__(self, idx):
		#get an entry
		key = list(self.data.keys())[idx]
		datum = self.data[key]
		datum_out = {}
		### JOINTS 3D
		joints_3d_raw = load_skel3D_bin(datum["joints_3d_raw"], scale= 1e-03)
		joints_3d = torch.cat([joints_3d_raw[:, :, joint_id, :] for joint_id in joint_selection] , dim=0).unsqueeze(0).permute(0,2,1,3)
		datum_out.update({"joints_3d" : joints_3d.squeeze(0).squeeze(0)})
		if self.mode == "single":
			#Calibration extrinsics
			rot, trans, inv_rot, inv_trans = load_extrinsics(\
				datum["pose"])
			if "color" in self.data_types and not "depth" in self.data_types:
				color_tensor = load_image(datum["color"])
				color_intrinsics, color_intrinsics_inv = get_intrinsics(\
					datum["dev_name"], self.color_intrinsics_repository, 1, self.center_crop)
				## DATUM UPDATE
				datum_out.update({
					"color" : color_tensor.squeeze(0),
					"cintr" : color_intrinsics,
					"campose_rotation" : inv_rot, 
					"campose_translation" :  inv_trans
					})
			elif "depth" in self.data_types and not "color" in self.data_types:				
				depth_tensor = load_depth_pgm(datum["depth"], scale = 1e-04)
				mask = (depth_tensor < self.depth_threshold).type(torch.float32)
				masked_depth_tensor = depth_tensor * mask
				depth_intrinsics, depth_intrinsics_inv = get_intrinsics(\
					datum["dev_name"], self.depth_intrinsics_repository, 4, self.center_crop)

				depth_intrinsics_down4, depth_intrinsics_inv_down4 = get_intrinsics(\
					datum["dev_name"], self.depth_intrinsics_repository, 16, self.center_crop)
				## DATUM UPDATE
				datum_out.update({
					"dintr" : depth_intrinsics,
					"dintr_down4" : depth_intrinsics_down4,
					"depth" : masked_depth_tensor.squeeze(0), 
					"campose_rotation" : inv_rot, 
					"campose_translation" :  inv_trans
					})
			elif "depth" in self.data_types and "color" in self.data_types:
				color_tensor = load_image(datum["color"])
				color_intrinsics, color_intrinsics_inv = get_intrinsics(\
					datum["dev_name"], self.color_intrinsics_repository, 1, self.center_crop)
				depth_tensor = load_depth_pgm(datum["depth"], scale = 1e-04)
				mask = (depth_tensor < self.depth_threshold).type(torch.float32)
				masked_depth_tensor = depth_tensor * mask
				depth_intrinsics, depth_intrinsics_inv = get_intrinsics(\
					datum["dev_name"], self.depth_intrinsics_repository, 4, self.center_crop)
				## DATUM UPDATE
				datum_out.update({
					"color" : color_tensor.squeeze(0),
					"cintr" : color_intrinsics,
					"dintr" : depth_intrinsics,
					"depth" : masked_depth_tensor.squeeze(0), 
					"campose_rotation" : inv_rot, 
					"campose_translation" :  inv_trans
					})
		else:
			for device in self.device_list:
				color_tensor = load_image(datum[device]["color"])
				color_intrinsics, color_intrinsics_inv = get_intrinsics(\
				device, self.color_intrinsics_repository, 1)

				depth_tensor = load_depth_pgm(datum[device]["depth"], scale = 1e-04)
				mask = (depth_tensor < self.depth_threshold).type(torch.float32)
				masked_depth_tensor = depth_tensor * mask
				depth_intrinsics, depth_intrinsics_inv = get_intrinsics(\
				device, self.depth_intrinsics_repository, 4)
				#Calibration extrinsics
				rot, trans, inv_rot, inv_trans = load_extrinsics(\
					datum[device]["pose"])
				## DATUM UPDATE
				datum_out.update({
					device + "_color" : color_tensor.squeeze(0),
					device + "_cintr" : color_intrinsics,
					device + "_dintr" : depth_intrinsics,
					device + "_depth" : masked_depth_tensor.squeeze(0), 
					device + "_campose_rotation" : inv_rot, 
					device + "_campose_translation" :  inv_trans,
					})
		
		return datum_out

	def get_data(self):
		return self.data


import moai.log.lightning as milog
import pytorch_lightning
import hydra.utils as hyu
import omegaconf.omegaconf
import typing
import logging

log = logging.getLogger(__name__)

__all__ = ["LightningPlayer"]

class LightningPlayer(pytorch_lightning.Trainer):
    def __init__(self,
        logging: omegaconf.DictConfig=None,
        num_nodes: int=1,
        gpus: typing.Optional[typing.Union[typing.List[int], str, int]]=None,
        tpu_cores: typing.Optional[int]=None,
        overfit_batches: float=0.0,
        fast_dev_run: bool=False,
        limit_train_batches: float=1.0,
        limit_val_batches: float=1.0,
        limit_test_batches: float=1.0,
        val_check_interval: float=1.0,
        distributed_backend: typing.Optional[str]=None,
        weights_summary: typing.Optional[str]='full',
        amp_level: str='O1',
        **kwargs
    ):
        logger = hyu.instantiate(logging) if logging is not None else milog.NoOp()
        super(LightningPlayer, self).__init__(
            logger=logger,      
            num_nodes=num_nodes,
            gpus=gpus,
            tpu_cores=tpu_cores,
            overfit_batches=overfit_batches,
            fast_dev_run=fast_dev_run,
            max_epochs=1,
            min_epochs=1,
            limit_train_batches=limit_train_batches,
            limit_val_batches=limit_val_batches,
            limit_test_batches=limit_test_batches,
            val_check_interval=val_check_interval,
            distributed_backend=distributed_backend,
            weights_summary=weights_summary,
            amp_level=amp_level,
            **kwargs
        )

    def play(self, model):
        self.fit(model)
        log.info("Train & validation data completed, continuing with test data.")
        self.test(model)


import torch

__all__ = [
    "Scalar",
    "Random",
    "Ones",
    "Zeros",
    "Clone",
]

class Scalar(torch.nn.Module):
    def __init__(self, 
        value: float
    ):
        super(Scalar, self).__init__()
        self.value = value
    
    def forward(self, tensor: torch.Tensor) -> torch.Tensor:
        return torch.scalar_tensor(self.value,
            dtype=tensor.dtype, device=tensor.device
        )

class Random(torch.nn.Module):
    def __init__(self):
        super(Random, self).__init__()

    def forward(self, tensor: torch.Tensor) -> torch.Tensor:
        # return torch.rand_like(tensor)
        return torch.rand(1, *tensor.shape[1:], 
            dtype=tensor.dtype, device=tensor.device).expand_as(tensor)

class Ones(torch.nn.Module):
    def __init__(self):
        super(Ones, self).__init__()

    def forward(self, tensor: torch.Tensor) -> torch.Tensor:
        # return torch.ones_like(tensor)
        return torch.ones(1, *tensor.shape[1:], dtype=tensor.dtype,
                device=tensor.device).expand_as(tensor) if tensor.shape\
            else torch.scalar_tensor(1, dtype=tensor.dtype, device=tensor.device)

class Zeros(torch.nn.Module):
    def __init__(self):
        super(Zeros, self).__init__()

    def forward(self, tensor: torch.Tensor) -> torch.Tensor:
        # return torch.zeros_like(tensor)
        return torch.zeros(1, *tensor.shape[1:], 
            dtype=tensor.dtype, device=tensor.device).expand_as(tensor)

class Clone(torch.nn.Module):
    def __init__(self):
        super(Clone, self).__init__()

    def forward(self, tensor: torch.Tensor) -> torch.Tensor:
        return tensor.clone()


import cv2
import torch
import numpy
import io

#NOTE: extract these to common loading funcs

def load_image(filename: str, data_type=torch.float32):
    color_img = numpy.array(cv2.imread(filename, cv2.IMREAD_ANYCOLOR))
    color_img =cv2.cvtColor(color_img, cv2.COLOR_BGR2RGB)
    h, w, c = color_img.shape
    color_data = color_img.astype(numpy.float32).transpose(2, 0, 1)
    return torch.from_numpy(
        color_data.reshape(1, c, h, w)        
    ).type(data_type) / 255.0

def load_depth(filename: str, data_type=torch.float32, scale=0.001):
    depth_img = numpy.array(cv2.imread(filename, cv2.IMREAD_ANYDEPTH))
    h, w = depth_img.shape
    depth_data = depth_img.astype(numpy.float32) * scale
    return torch.from_numpy(
        depth_data.reshape(1, 1, h, w)        
    ).type(data_type)

def crop_depth(filename: str, data_type=torch.float32, scale=0.001):
    depth_img = numpy.array(cv2.imread(filename, cv2.IMREAD_ANYDEPTH))
    center_cropped_depth_img = depth_img[60:420, 0:640]
    h, w = center_cropped_depth_img.shape
    depth_data = center_cropped_depth_img.astype(numpy.float32) * scale
    return torch.from_numpy(
        depth_data.reshape(1, 1, h, w)        
    ).type(data_type)

def readpgm(name: str): #NOTE: for ascii PGM
    with open(name) as f:
        lines = f.readlines()
    # Ignores commented lines
    for l in list(lines):
        if l[0] == '#':
            lines.remove(l)
    # Makes sure it is ASCII format (P2)
    assert lines[0].strip() == 'P2' 
    # Converts data to a list of integers
    data = []
    for line in lines[1:]:
        data.extend([int(c) for c in line.split()])

    buffer = numpy.array(data[3:])
    return buffer.reshape(data[1], data[0])


def load_depth_pgm(filename: str, data_type=torch.float32, scale = 1):
    depth_img = readpgm(filename)
    depth_img = depth_img.astype(numpy.float32) * scale
    h, w = depth_img.shape
    depth_data = depth_img.astype(numpy.float32)
    return torch.from_numpy(
        depth_data.reshape(1, 1, h, w)        
    ).type(data_type)
    
# define a function which returns an image as numpy array from figure
def get_img_from_fig(fig, dpi=180):
    buf = io.BytesIO()
    fig.savefig(buf, format="png", dpi=dpi)
    buf.seek(0)
    img_arr = numpy.frombuffer(buf.getvalue(), dtype=numpy.uint8)
    buf.close()
    img = cv2.imdecode(img_arr, 1)
    # img = cv2.cvtColor(img, cv2.COLOR_BGR2RGB)
    return torch.from_numpy(
        img.reshape(1, img.shape[2], img.shape[0], img.shape[1])        
    ).type(torch.float32) / 255.0


import csv
import itertools
import cv2
import numpy
import torch
import torchvision
import enum
import logging
import typing

__all__ = [
    "Placements",
    "ImageTypes",
    "_load_paths",
    "_rotate_image",
    "_filename_separator",
    "_rotate_normal_image",
    "extract_image",
    "extract_path",
]

class Placements(enum.Enum):
    CENTER = 1    
    RIGHT = 2
    UP = 3

    def describe(self):        
        return self.name.lower(), self.value

    def __str__(self) -> str:
        n = str.lower(self.name) if self.value != 1 else str("left_down")
        return n

class ImageTypes(enum.Enum):
    COLOR = 1    
    DEPTH = 2
    NORMAL = 3

    def describe(self):        
        return self.name.lower(), self.value

    def __str__(self) -> str:
        return str.lower(self.name)        

_dataset_images_order = [
    str.format("{}_{}", str(Placements.CENTER), str(ImageTypes.COLOR)),
    str.format("{}_{}", str(Placements.RIGHT), str(ImageTypes.COLOR)),
    str.format("{}_{}", str(Placements.UP), str(ImageTypes.COLOR)),
    str.format("{}_{}", str(Placements.CENTER), str(ImageTypes.DEPTH)),
    str.format("{}_{}", str(Placements.RIGHT), str(ImageTypes.DEPTH)),
    str.format("{}_{}", str(Placements.UP), str(ImageTypes.DEPTH)),
    str.format("{}_{}", str(Placements.CENTER), str(ImageTypes.NORMAL)),
    str.format("{}_{}", str(Placements.RIGHT), str(ImageTypes.NORMAL)),
    str.format("{}_{}", str(Placements.UP), str(ImageTypes.NORMAL)),
]

_filename_separator = "_path"

def _create_selectors(
    placements: typing.Sequence[Placements],
    types: typing.Sequence[ImageTypes],
) -> typing.List[str]:
    return list(map( \
        lambda o: any(str(p) in o for p in placements) and any(str(t) in o for t in types), \
            _dataset_images_order))

def _load_paths(
    filename: str,
    name: str,
    placements: typing.Sequence[Placements],
    image_types: typing.Sequence[ImageTypes],
) -> typing.Dict[str, typing.Dict[str, str]]:
    selectors = _create_selectors(placements, image_types)
    with open (filename, 'r') as f:
        rows = [list(itertools.compress(row, selectors)) for row in csv.reader(f, delimiter=' ')]
        return list(map(
            lambda row: dict(list(
                (
                    str(p), dict(list(
                    (                        
                        str(t) + _filename_separator, \
                            next(filter(lambda r: str(p) in r.lower() and str(t) in r.lower(), row))
                    ) for t in image_types
                    ))
                ) for p in placements
            ))
        , [row for row in rows if all(map(lambda r: name in r, row))]
        ))

def _load_color_image( #TODO: extract these to a common module
    filename: str,
    data_type: torch.dtype
)  -> torch.Tensor:
    return torchvision.io.read_image(filename) / 255.0

def _load_depth_image( #TODO: extract these to a common module
    filename: str,
    data_type: torch.dtype
)  -> torch.Tensor:
    depth_img = numpy.array(cv2.imread(filename, cv2.IMREAD_ANYDEPTH))
    h, w = depth_img.shape
    return torch.from_numpy(depth_img).type(data_type).reshape(1, h, w)

def _load_normal_image( #TODO: extract these to a common module
    filename: str,
    data_type: torch.dtype
)  -> torch.Tensor:#TODO: flip y-z channels
    normal_img = numpy.array(cv2.imread(filename, cv2.IMREAD_UNCHANGED)).transpose(2, 0, 1)
    c, h, w = normal_img.shape
    return torch.from_numpy(normal_img).reshape(c, h, w).type(data_type)

def _rotate_image(image: torch.Tensor, idx: int) -> torch.Tensor:
    width = image.shape[2]
    if idx < 0 or idx >= width:
        return
    rotated = torch.roll(image, idx, 2)
    return rotated

def _get_rotation_matrix(idx: int, width: int, dim: int=2) -> torch.Tensor:
    theta = (float(idx) / float(width)) * 2.0 * numpy.pi
    mat = [
        [1.0,       0.0,              0.0],\
        [0.0,       numpy.cos(theta),    numpy.sin(theta)],\
        [0.0,       numpy.sin(theta),    numpy.cos(theta)]]
    mat = numpy.asarray(mat, dtype = numpy.float32)
    mat = mat.transpose()
    return torch.as_tensor(torch.from_numpy(mat), dtype = torch.float32)

def _rotate_normal_image(image: torch.Tensor, idx: int) -> torch.Tensor:
    width = image.shape[2]
    if idx < 0 or idx >= width:
        return
    rot_mat = _get_rotation_matrix(idx, width)
    tmp_img = image.permute((1, 2, 0))
    rot_img = torch.matmul(tmp_img, rot_mat)
    rot_img = rot_img.permute((2, 0, 1))
    rot_img = torch.roll(rot_img, idx, 2)
    return rot_img
    
_image_loaders = {
    "color": lambda *params: _load_color_image(params[0], params[1]),
    "depth": lambda *params: _load_depth_image(params[0], params[1]),
    "normal": lambda *params: _load_normal_image(params[0], params[1]),
}

def _load_image(
    filename: str,
    image_type: ImageTypes,
    data_type: torch.dtype=torch.float32
) -> torch.Tensor:
    return _image_loaders[image_type](filename, data_type)

def extract_image(
    tensor_dict: typing.Dict[str, torch.Tensor],
    placement: Placements,
    image_type: ImageTypes
) -> torch.Tensor:
    if not str(placement) in tensor_dict or not str(image_type) in tensor_dict[str(placement)]:
        logging.fatal("Could not extract the requested placement/image (%s/%s) from the given tensor dictionary." % (placement, image_type))
    return tensor_dict[str(placement)][str(image_type)]

def extract_path(
    tensor_dict: typing.Dict[str, torch.Tensor],
    placement: Placements,
    image_type: ImageTypes
) -> str:
    if not str(placement) in tensor_dict or not str(image_type) + _filename_separator in tensor_dict[str(placement)]:
        logging.fatal("Could not extract the requested placement/image (%s/%s) path from the given tensor dictionary." % (placement, image_type))
    return tensor_dict[str(placement)][str(image_type) + _filename_separator]


from moai.action.train import train
from moai.action.evaluate import evaluate
from moai.action.play import play
from moai.action.diff import diff
from moai.action.plot import plot

import omegaconf.omegaconf
import hydra
import logging
import sys
import os
import pkgutil

from rich.traceback import install
install(width=120, extra_lines=5, theme=None,
    word_wrap=True, show_locals=False, indent_guides=True,    
)

ERROR_FORMAT = "%(levelname)s at %(asctime)s in %(funcName)s in %(filename) at line %(lineno)d: %(message)s"
DEBUG_FORMAT = "%(lineno)d in %(filename)s at %(asctime)s: %(message)s"
# FORMAT = "[%(asctime)s][%(filename)s][%(levelname)s] - %(message)s"
FORMAT = "[%(levelname)s] - %(message)s"

logging.basicConfig(level=logging.INFO, format=FORMAT)
log = logging.getLogger('moai')

def debug(cfg):
    log.info(cfg)

__MODES__ = {
    'train': train,
    'evaluate': evaluate,
    'play': play,
    'diff': diff,
    'plot': plot,
    'debug': debug,
}

__MIN_ARGS_COUNT__ = {
    'train': 2,
    'evaluate': 2,
    'play': 2,
    'diff': 1,
    'plot': 1,
    'debug': 2,
}

def run(cfg: omegaconf.DictConfig):
    __MODES__[cfg.mode](cfg)

def moai():    
    # os.environ['HYDRA_FULL_ERROR'] = '1'
    mode = sys.argv.pop(1)
    if mode not in __MODES__:
        log.error(f"Wrong moai execution mode ({mode}), supported modes are: {list(__MODES__.keys())}.")
        sys.exit(-1)
    min_args = __MIN_ARGS_COUNT__[mode]
    if len(sys.argv) < min_args:
        log.error(f"Insufficient arguments provided for moai. Calling should specify the mode and config file: \'moai MODE CONFIG\'.")    
    config = sys.argv.pop(1) if min_args > 1 else f"tools/{mode}.yaml" 
    other_args = []
    if min_args > 1 or mode == 'plot':
        output_dir = "hydra.run.dir=actions/" + mode + "/${now:%Y-%m-%d}/${now:%H-%M-%S}-${experiment.name}"
        other_args.append(output_dir)
    else:
        other_args.append("+hydra.hydra_logging.root.handles=[]")
        other_args.append("hydra.hydra_logging.disable_existing_loggers=true")
        other_args.append("hydra.output_subdir=null")
    for oarg in other_args:
        sys.argv.append(oarg)
    sys.argv.append(f"+mode={mode}")
    main = hydra.main(config_path="conf", config_name=config)(run)
    main()

if __name__ == "__main__":
    moai()


import numpy
import torch

def load_skel3D_bin(
    filename: str,
    data_type = torch.float32,
    scale: float=1.0
):
    skel = numpy.load(filename).astype(numpy.float32) * scale
    num_people, num_joints, coords = skel.shape    
    return torch.from_numpy(
        skel.reshape(1, num_people, num_joints, coords)
    ).type(data_type)


from moai.supervision.losses.regression.robust.berhu import Berhu
from moai.supervision.losses.regression.robust.welsch import Welsch
from moai.supervision.losses.regression.robust.wing import Wing

__all__ = [
    "Berhu",
    "Welsch",
    "Wing",
]


import json
import numpy
import torch

__all__ = [
    'load_intrinsics_repository',
    'get_intrinsics',
]

def load_intrinsics_repository(filename: str, stream: str):    
    '''
        use color for loading color intrinsics and depth for depth intrinsics
    '''
    with open(filename, 'r') as json_file:
        intrinsics_repository = json.load(json_file)
        if stream == "depth":
            intrinsics_dict = dict((intrinsics['Device'], \
                intrinsics['Depth Intrinsics'][0]['1280x720'])\
                    for intrinsics in intrinsics_repository)
        if stream == "color":
            intrinsics_dict = dict((intrinsics['Device'], \
                intrinsics['Depth Intrinsics'][0]['1280x720'])\
                    for intrinsics in intrinsics_repository)
    return intrinsics_dict

def get_intrinsics(name, intrinsics_dict, scale=1, center_crop={'width': 0, 'height': 0}, data_type=torch.float32):
    if intrinsics_dict is not None:
        intrinsics_data = numpy.array(intrinsics_dict[name])
        intrinsics = torch.tensor(intrinsics_data).reshape(3, 3).type(data_type)   
        intrinsics[0, 2] = intrinsics[0, 2] - center_crop['width']
        intrinsics[1, 2] = intrinsics[1, 2] - center_crop['height']

        intrinsics[0, 0] = intrinsics[0, 0] / scale
        intrinsics[0, 2] = intrinsics[0, 2] / scale
        intrinsics[1, 1] = intrinsics[1, 1] / scale
        intrinsics[1, 2] = intrinsics[1, 2] / scale
        intrinsics_inv = intrinsics.inverse()
        return intrinsics, intrinsics_inv
    raise ValueError("Intrinsics repository is empty")    


joint_selection = [
    0,  # Hips
    2,  # Spine1
    5,  # Neck
    8,  # Head
    10, # RightArm
    11, # RightForeArm
    12, # RightHand
    16, # LeftArm
    17, # LeftForeArm
    18, # LeftHand
    21, # RightUpLeg
    22, # RightLeg
    23, # RightFoot
    25, # RightToeBase
    27, # LeftUpLeg
    28, # LeftLeg
    29, # LeftFoot
    31, # LeftToeBase
]


from moai.data.datasets.spherical.ThreeD60.common import (
    _load_paths,
    _load_image,
    _filename_separator,
    Placements,
    ImageTypes,
)

import os
import copy
import logging
import torch
import typing

log = logging.getLogger(__name__)

__all__ = ["Matterport3D"]

class Matterport3D(torch.utils.data.Dataset):
    def __init__(self, 
        filename: str,
        placements: typing.Sequence[Placements],
        image_types: typing.Sequence[ImageTypes],
    ):
        assert(os.path.exists(filename))
        super(Matterport3D, self).__init__()
        self.entries = _load_paths(filename, "Matterport", placements, image_types)
        log.info(f"Loaded {len(self.entries)} items from Matterport3D.")

    def __len__(self) -> int:
        return len(self.entries)

    def __getitem__(self, index: int) -> typing.Dict[str, torch.Tensor]:
        output = { }
        for placement, type_map in self.entries[index].items():
            for typed_path, filename in type_map.items():
                image_type = typed_path.replace(_filename_separator, "")
                output[placement + "_" + image_type] = _load_image(filename, image_type)
        return output